unit UExpr;
interface
      (* Unit pour elavuler une expression arithmtique  logique booleenne
        *
        *              ar montor
        *
        *
        *)

uses
    SysUtils,Variants,Dialogs, UHashStr,UPasParser;
type
    TIdentType=(itFunc,itVar);
    PExpFunc = ^TExpFunc;
    PArgList = ^TArgList;
    TArgList = array[Byte]of Variant;
    TExpFunc = function:Variant of object;
    PIdentInfo =^TIdentInfo;
    TIdentInfo = packed record
      NextItem:PIdentInfo;
      IdentType:TIdentType;
      case  integer of
         0:( ParamsCount:integer;
             ExpFunc:PExpFunc);
         1:(VarAddr:PVariant)
    end;


    TCustomExpEval=class
    private
      FParser :TPasParser;
      FIdenList :TStrHashList;
      FPtrList :PIdentInfo;
      FStack:array of Variant;
      FIdx:integer;
      FArgs: PArgList;
      FArgsCount: integer;
      FRetToken :TokenType;
      FReturn :Variant;
      function GetParserPos: integer;
      function Priority(Op: TOperatorKind):integer;
      function GetFactor: TokenType;
      procedure GetTerme(Op: TOperatorKind);
      procedure GetArgList(Opener,Closer:TokenType);
   protected
      function FindIden(const Ident:string):PIdentInfo;
      function Pop(ALen:integer=1):Variant;
      function AsNumber():Variant;
      function NewIdentInfo(const Ident:string):PIdentInfo;
      function AddFunc(const Ident:string;FuncExp:TExpFunc;ParamsCount:integer = -1):PIdentInfo;
      function AddVar(const Ident:string;const Value:Variant):PIdentInfo;
      function NextToken():TokenType;
      function TryNextToken(tok: TokenType):boolean;
      property ArgsCount:integer read FArgsCount;
      procedure ClearIdent;
      procedure SynCheck(eStatus: Boolean=False);
      procedure PushSys();
      procedure PopSys();
      procedure Push(const F:Variant);
      procedure AsString();
      procedure GetExpression;
      procedure GetInstruction;
      procedure UProc(Op: TOperatorKind);
      procedure BProc(Op: TOperatorKind);
      procedure IDentProc();
      property Args:PArgList read FArgs;
    public
      constructor Create();
      destructor Destroy();override;
      function EvalExp(const Value: string):Variant;
      property ParserPos:integer read GetParserPos;
   end;

   TExpEval=class(TCustomExpEval)
   protected
      function FloatFmt():Variant;
      function Avg():Variant;
      function Pi():Variant;
   public
      constructor Create();
   end;

   function Eval(const ExpTxt:string):Variant;
implementation

uses math;

const STACKSIZE = 1024;


function Eval(const ExpTxt:string):Variant;
begin
  with TExpEval.Create do
  try
        Result := EvalExp(ExpTxt);
  finally
    Free();
  end;
end;

{ TCustomExpEval }

constructor TCustomExpEval.Create;
begin
   FIdenList :=TStrHashList.Create();// Liste des noms fonctions variables
   FParser :=TPasParser.Create;
   SetLength(FStack,STACKSIZE);
end;

destructor TCustomExpEval.Destroy;
begin
  ClearIdent();
  FIdenList.Free;
  FParser.Free;
  inherited;
end;

function TCustomExpEval.NextToken: TokenType;
begin
    if FRetToken = ttNone then
    begin
       Result :=FParser.NextToken;
       SynCheck(Result <> ttInvalid);
    end else  begin
       Result :=FRetToken;
       FRetToken := ttNone;
    end;
end;

function TCustomExpEval.TryNextToken(tok: TokenType): boolean;
begin
    FRetToken :=NextToken;
    Result:=FRetToken = tok;
    if Result then
       FRetToken := ttNone;
end;

function TCustomExpEval.GetParserPos: integer;
begin
   Result:=FParser.Position;
end;

procedure TCustomExpEval.SynCheck(eStatus: Boolean);
begin
    if not eStatus then
       raise Exception.CreateFmt('Erreur de syntaxe a la position : %d',
          [ParserPos]);
end;

function TCustomExpEval.FindIden(const Ident: string): PIdentInfo;
var
 Idx :integer;
begin
     Idx := FIdenList.ValueOf(Uppercase(Ident));
     if Idx = -1 then
       Result := nil
     else
       Result:=PIdentInfo(Idx);
end;

function TCustomExpEval.NewIdentInfo(const Ident: string): PIdentInfo;
begin
    if Assigned(FindIden(Ident))then
       raise Exception.Create('Identificateur existe');
    New(Result);
    Result.NextItem := FPtrList;
    FPtrList := Result;
    FIdenList.Add(Uppercase(Ident),Integer(Result));
end;

function TCustomExpEval.AddFunc(const Ident: string; FuncExp: TExpFunc;ParamsCount:integer = -1):PIdentInfo;
begin
    Result :=NewIdentInfo(Ident);
    Result.IdentType :=itFunc;
    Result.ParamsCount :=ParamsCount;
    New(Result.ExpFunc);
    Result.ExpFunc^:=FuncExp;
end;

function TCustomExpEval.AddVar(const Ident: string; const Value: Variant):PIdentInfo;
begin
    Result :=NewIdentInfo(Ident);
    Result.IdentType :=itVar;
    New(Result.VarAddr);
    Result.VarAddr^:=Value;
end;

procedure TCustomExpEval.ClearIdent();
var
 P,Tmp:PIdentInfo;
begin
   P:=FPtrList;
   while  P <> nil do
   begin
      Tmp :=P.NextItem;
      with P^ do
      case IdentType of
           itFunc : Dispose(ExpFunc);
           itVar  : Dispose(VarAddr);
      end;
      Dispose(P);
      P :=Tmp;
   end;
   FPtrList:=nil;
end;

function TCustomExpEval.EvalExp(const Value: string):Variant;
begin
    FRetToken := ttNone;
    FIdx := -1;
    FParser.Code := Value;
    Push(0);
    GetInstruction();
    Result:=Pop;
    if FIdx <> -1 then
      raise Exception.Create('Pile corrompue');
end;

function TCustomExpEval.Pop(ALen:integer): Variant;
begin
    while ALen > 0 do
    begin
      Result:=FStack[FIdx];
      FStack[FIdx] :=Null;
      dec(FIdx);
      dec(ALen);
    end;
end;

procedure TCustomExpEval.Push(const F: Variant);
begin
   inc(FIdx);
   FStack[FIdx]:=F;
end;

function TCustomExpEval.AsNumber: Variant;
begin
    try
       Result:=StrToInt(FParser.TokenTxt);
    except
       Result:=StrToFloat(FParser.TokenTxt);
    end;
    Push(Result);
end;

procedure TCustomExpEval.AsString;
var
  S:string;
begin
     S := Copy(FParser.TokenTxt,2,Length(FParser.TokenTxt)-2);
     S := StringReplace(S,'""','"',[rfReplaceAll]);
     Push(S);
end;

function TCustomExpEval.GetFactor():TokenType;
var
 UOp:TOperatorKind;
begin
    Result:=NextToken;
    UOp:=tkOpPos;
    if Result = ttUOp then
    begin
       UOp:=FParser.Operator;
       Result:=NextToken;
    end;
    case Result of
       ttNumber : AsNumber();
       ttIdent  : IDentProc();
       ttString : AsString();
       ttOpenExp:begin
                    GetExpression();
                    SynCheck(FParser.CurrToken = ttCloseExp );
                 end;
       else
        SynCheck();
    end;
    if UOp <> tkOpPos then
      UProc(UOp);
    Result:=NextToken;
end;

procedure TCustomExpEval.GetTerme(Op: TOperatorKind);
begin
    if GetFactor() = ttOp then
          while (FParser.CurrToken=ttOp)and (Priority(FParser.Operator)<Priority(Op)) do
               GetTerme(FParser.Operator);

    BProc(Op)
end;

procedure TCustomExpEval.GetExpression();
begin
     if GetFactor() = ttOp then
         while (FParser.CurrToken=ttOp) do
                GetTerme(FParser.Operator);
end;

procedure TCustomExpEval.GetInstruction;
var
 pIdnt:PIdentInfo;
 DoExit:boolean;
begin
   DoExit :=False;
   repeat
         case NextToken of
             ttCmd:begin
                        case FParser.Cmd  of
                             ccReturn :
                                     begin
                                         Pop;
                                         GetExpression();
                                         DoExit :=True;
                                     end;
                             ccEcho :
                                     begin
                                         GetExpression();
                                         Showmessage(Pop);
                                     end;
                          end;
                          SynCheck( FParser.CurrToken in [ttEndInst,ttNone]);
                          NextToken;
                    end;
             ttIdent:
               begin
                   pIdnt:= FindIden(FParser.TokenTxt);
                   if pIdnt = nil then
                      pIdnt :=AddVar(FParser.TokenTxt,0);
                   case pIdnt.IdentType of
                        itFunc : begin
                                     IDentProc();
                                     Pop;
                                     NextToken
                                 end;
                        itVar  :begin
                                     NextToken;
                                     SynCheck(FParser.CurrToken = ttAssign);
                                     GetExpression();
                                     pIdnt.VarAddr^:= Pop;
                               end;
                   end;

              end;
            // else SynCheck();
         end;
        SynCheck( FParser.CurrToken in [ttEndInst,ttNone]);
  until  DoExit or (FParser.CurrToken = ttNone);

end;

function TCustomExpEval.Priority(Op: TOperatorKind): integer;
begin
   case Op of
        tkOpPow:          Result:=1;
        tkOpMul..tkOpShr :Result:=2;
        tkOpAdd..tkOpXor :Result:=3;
        tkOpEqu..tkOpGreatEqu :Result:=4;
        else
          Result:=0;
   end;
end;

procedure TCustomExpEval.BProc(Op: TOperatorKind);
var
  A1,A2,Result:Variant;
begin
  A2:=Pop;
  A1:=Pop;
  case Op of
      tkOpMul:  Result:= A1 *   A2;
      tkOpDivF: Result:= A1 /   A2;
      tkOpShl:  Result:= A1 shl A2;
      tkOpShr:  Result:= A1 shr A2;
      tkOpMod:  Result:= A1 mod A2;
      tkOpDiv:  Result:= A1 div A2;
      tkOpAnd:  Result:= A1 and A2;
      tkOpAdd:  Result:= A1 +   A2;
      tkOpSub:  Result:= A1 -   A2;
      tkOpOr:   Result:= A1 or  A2;
      tkOpXor:  Result:= A1 xor A2;
      tkOpPow:  Result:=Power(A1 , A2);
      tkOpEqu:  Result:=A1 = A2;
      tkOpLess: Result:=A1 < A2;
      tkOpGreat:    Result:=A1 > A2;
      tkOpLessEqu:  Result:=A1 <= A2;
      tkOpNotEqu:   Result:=A1 <> A2;
      tkOpGreatEqu: Result:=A1 >= A2;
      else
          Result:= 0;
  end;
    Push(Result);

end;

procedure TCustomExpEval.UProc(Op: TOperatorKind);
var
  Result:Variant;
begin
  case Op of
    tkOpNeg: Result:= -Pop;
    tkOpNot: Result:= not Pop;
    tkRound: Result:= Round(Pop);
    tkSqrt:  Result:= Sqrt(Pop);
    tkSqr:   Result:= Sqr(Pop);
    tkInt:   Result:= Int(Pop);
    tkFrac:  Result:= Frac(Pop);
    tkAbs:   Result:= Abs(Pop);
    tkSin:   Result:= Sin(Pop);
    tkCos:   Result:= Cos(Pop);
    tkExp:   Result:= Exp(Pop);
    tkArctan:Result:= Arctan(Pop);
    tkLn:    Result:= Ln(Pop);
    tkOpPos: Result:= Pop;
    else
       Result:= 0;
  end;
  Push(Result);
end;

procedure TCustomExpEval.PushSys;
begin
       Push(integer(FArgs));
       Push(FArgsCount);
       FArgsCount := 0;
       FArgs := @FStack[FIdx+1];
end;

procedure TCustomExpEval.PopSys;
begin
       Pop(FArgsCount);
       FArgsCount :=Pop;
       integer(FArgs) :=Pop;
       Push(FReturn);
       FReturn :=NULL;
end;

procedure TCustomExpEval.GetArgList(Opener,Closer:TokenType);
begin
       if Opener = ttNone then
       begin
             repeat
                 GetExpression;
                 inc(FArgsCount);
             until FParser.CurrToken<>ttSep;
       end
       else if TryNextToken(Opener) then begin
            if TryNextToken(Closer) then
               Exit;
            repeat
               GetExpression;
               inc(FArgsCount);
            until FParser.CurrToken<>ttSep;
            SynCheck(FParser.CurrToken = Closer);
       end;
end;

procedure TCustomExpEval.IDentProc();
var
  pIdenInf:PIdentInfo;
begin
   pIdenInf:= FindIden(FParser.TokenTxt);
   SynCheck(pIdenInf <> nil);
   with pIdenInf^ do
   case IdentType  of
      itVar : Push(VarAddr^);
      itFunc:
             begin
                 PushSys();
                 GetArgList(ttOpenExp,ttCloseExp);
                 if (ParamsCount <> -1)and ( ParamsCount <> ArgsCount) then
                    raise Exception.Create('Params count error');
                 FReturn := ExpFunc^();
                 PopSys();
             end;
   end;
end;


{ TExpEval }
constructor TExpEval.Create;
begin
  inherited;
  AddFunc('FloatFmt',FloatFmt,2);
  AddFunc('Avg',Avg);
  AddFunc('Pi',Self.Pi,0);
end;

function TExpEval.FloatFmt: Variant;
begin
    Result := Format(Args[0],[Double(Args[1])]);
end;

function TExpEval.Avg: Variant;
var
 I:integer;
begin
    if ArgsCount = 0 then
       raise Exception.Create('');
    Result := 0;
    for I := 0 to ArgsCount -1 do
           Result := Result+Args[I];
    Result := Result / ArgsCount;
end;


function TExpEval.Pi: Variant;
begin
   Result := 3.1415926535897932384626433832795
end;

initialization
   DecimalSeparator:= '.';
end.