unit UPasParser;
      (*  TPasParser parseur de code par Montor *)
interface

uses
  SysUtils,UHashStr;
   type
    TokenType=(ttNone,ttNumber,ttString,ttIdent,ttKeyWord,ttCmd,ttOpenExp,
                ttCloseExp,ttUOp,ttOp,ttSep,ttDef,ttEndInst,ttAssign,ttInvalid);
    TOperatorKind= (
        tkOpPos,tkOpNeg,tkOpNot,tkOpPow,tkOpMul,tkOpDivF, tkOpDiv, tkOpMod,
        tkOpAnd, tkOpShl,tkOpShr,tkOpAdd,tkOpSub,tkOpOr,tkOpXor,
        tkOpEqu,tkOpLess,tkOpLessEqu,tkOpNotEqu,tkOpGreat,tkOpGreatEqu,
        //Exten
        tkAbs,tkSin,tkCos,tkArctan,tkExp,tkSqrt,tkSqr,tkLn,
        tkRound,tkInt,tkFrac
        );

    TKeyWords=(kwVar,kwConst);

     TCustomCommand=(ccEcho,ccReturn);

    TGroupRec=packed record
    case integer of
       0:(Hash :integer);
       1:(Group:TokenType;
          case integer of
           0:(Operator:TOperatorKind);
           1:(KWord:TKeyWords);
           2:(Cmd:TCustomCommand)
          );
    end;

    TPasParser=class
    private
      FCode :string;
      FWords :TStrHashList;
      FPos :integer;
      FCursor:integer;
      FCopyStart:integer;
      FCurrToken: TokenType;
      FOperator: TOperatorKind;
      FKeyWord: TKeyWords;
      FCmd: TCustomCommand;
      function GetTokenTxt():string;
      procedure AddBinOp(const Ident:string;AKey:TOperatorKind);
      procedure AddUnOp(const Ident:string;AKey:TOperatorKind);
      procedure AddKeyW(const Ident:string;AKey:TKeyWords);
      procedure EndTokenPos(AChars: TSysCharSet;AExclude:boolean =True);
      procedure SetCode(const Value: string);
      function GetIdentGroup: TokenType;
      function GetOperatorGroup: TokenType;
      procedure AddCmd(const Ident: string; AKey: TCustomCommand);
   public
      constructor Create();
      destructor Destroy();override;
      function NextToken():TokenType;
      property CurrToken:TokenType read FCurrToken;
      property Operator:TOperatorKind read FOperator;
      property KeyWord: TKeyWords read FKeyWord;
      property Cmd: TCustomCommand read FCmd;
      property Code:string read FCode write SetCode;
      property TokenTxt:string read GetTokenTxt;
      property Position:integer read FCopyStart;
    end;

implementation
{TPasParser}

constructor TPasParser.Create;
begin

    FWords :=TStrHashList.Create();
    AddBinOp('div',tkOpDiv); AddBinOp('mod',tkOpMod);AddBinOp('or',tkOpOr);
    AddBinOp('and',tkOpAnd); AddBinOp('shl',tkOpShl); AddBinOp('shr',tkOpShr);
    AddBinOp('xor',tkOpXor);

    AddUnOp('not',tkOpNot);
    AddUnOp('abs',tkAbs);   AddUnOp('sin',tkSin);   AddUnOp('cos',tkCos);
    AddUnOp('exp',tkExp);   AddUnOp('sqrt',tkSqrt); AddUnOp('int',tkInt);
    AddUnOp('sqr',tkSqr);   AddUnOp('ln',tkLn);     AddUnOp('frac',tkFrac);
    AddUnOp('round',tkRound); AddUnOp('arctan',tkArctan);

    AddKeyW('var',kwVar);  AddKeyW('const',kwConst);

    AddCmd('echo',ccEcho);
    AddCmd('return',ccReturn);

end;

destructor TPasParser.Destroy;
begin
  FWords.Free;
  inherited;
end;
procedure TPasParser.AddUnOp(const Ident: string; AKey: TOperatorKind);
var
 N:TGroupRec;
begin
  if FWords.ValueOf(UpperCase(Ident))= -1 then
  begin
     N.Group :=ttUOp;
     N.Operator :=AKey;
     FWords.Add(UpperCase(Ident),N.Hash);

  end;
end;

procedure TPasParser.AddBinOp(const Ident: string; AKey: TOperatorKind);
var
 N:TGroupRec;
begin
  if FWords.ValueOf(UpperCase(Ident))= -1 then
  begin
     N.Group :=ttOp;
     N.Operator :=AKey;
     FWords.Add(UpperCase(Ident),N.Hash);

  end;
end;

procedure TPasParser.AddCmd(const Ident: string; AKey: TCustomCommand);
var
 N:TGroupRec;
begin
  if FWords.ValueOf(UpperCase(Ident))= -1 then
  begin
     N.Group :=ttCmd;
     N.Cmd :=AKey;
     FWords.Add(UpperCase(Ident),N.Hash);

  end;
end;
procedure TPasParser.AddKeyW(const Ident: string; AKey: TKeyWords);
var
 N:TGroupRec;
begin
  if FWords.ValueOf(UpperCase(Ident))= -1 then
  begin
     N.Group :=ttKeyWord;
     N.KWord :=AKey;
     FWords.Add(UpperCase(Ident),N.Hash);
  end;
end;

function TPasParser.GetOperatorGroup: TokenType;
const OPRTYPE :array[boolean,boolean]of TOperatorKind =((tkOpAdd,tkOpPos),(tkOpSub,tkOpNeg));
     procedure IIf(C :Char;IfTrue,Ifalse:TOperatorKind);
     begin
         FOperator := Ifalse;
         if FCode[FPos+1]<> C then
             Exit;
         inc(FCursor);
         FOperator  := IfTrue;
     end;
begin
          Result:=ttOp;
          case FCode[FPos] of
             '+','-':begin
                       if not (FCurrToken in[ttNumber,ttIdent,ttCloseExp]) then
                          Result:=ttUOp;
                        FOperator:= OPRTYPE[FCode[FPos]='-', Result=ttUOp];
                     end;
             '=': FOperator := tkOpEqu;
             '*': FOperator := tkOpMul;
             '/': FOperator := tkOpDivF;
             '^': FOperator := tkOpPow;
             '>': IIf('=',tkOpGreatEqu,tkOpGreat);
             '<': begin
                    IIf('=',tkOpLessEqu,tkOpGreat);
                    IIf('>',tkOpNotEqu,tkOpGreat);
                  end;

          end;
end;

function TPasParser.GetIdentGroup: TokenType;
var
 Tmp:TGroupRec;
begin
     EndTokenPos(['a'..'z','A'..'Z','0'..'9','_']);
     Tmp.Hash:=FWords.ValueOf(UpperCase(TokenTxt));
     with Tmp  do
       if (Hash <> -1) then
       begin
           Result    := Group;
           FKeyWord  := KWord;
           FOperator := Operator;
           FCmd      := Cmd;
       end
         else
           Result := ttIdent;
end;

function TPasParser.NextToken: TokenType;
begin
   if FPos > Length(FCode)then
   begin
      Result := ttNone;
      FCurrToken :=Result;
      Exit;
   end;

   while FCode[FPos] in [#1..#32]do
         inc(FPos);
   FCursor   := FPos;
   FCopyStart:= FPos;
   case FCode[FPos] of
        'a'..'z',
        'A'..'Z','_': Result:=GetIdentGroup();

        '0'..'9':begin
                    Result:=ttNumber;
                    EndTokenPos(['0'..'9','.','e','E']);
                 end;
        '"':    begin
                    Result := ttString;
                    repeat
                       EndTokenPos(['"'],false);
                       inc(FCursor);
                       if FCode[FCursor+1]<> '"' then
                          break;
                       inc(FCursor);
                    until False;
                 end;
        '+','-','*','/',
        '^','<','>','=':Result := GetOperatorGroup();

        ';': Result := ttEndInst;
        '(': Result := ttOpenExp;
        ')': Result := ttCloseExp;
        ',': Result := ttSep;
        ':': begin
                Result := ttInvalid;
                if FCode[FPos+1]= '=' then
                begin
                   Result := ttAssign;
                   inc(FCursor);
                end;
             end;
        #0: Result := ttNone;
        else
          Result := ttInvalid;
  end;

  if (Result<>ttNone )then
      FPos:=FCursor+1;

   FCurrToken :=Result;
end;

procedure TPasParser.EndTokenPos(AChars:TSysCharSet;AExclude:boolean);
var
 I:integer;
begin
  for I:=FCursor to length(FCode)-1 do
  begin
       if(FCode[I+1]in AChars)xor AExclude then
          break;
       inc(FCursor);
  end;
end;

function TPasParser.GetTokenTxt: string;
begin
    Result:= Copy(FCode,FCopyStart,FCursor-(FCopyStart-1)) ;
end;

procedure TPasParser.SetCode(const Value: string);
begin
    FPos:=1;
    FCode := Value;
    FCurrToken := ttNone;
end;


end.
